from itertools import filterfalse, product
import numpy as np
from clease.cluster_fingerprint import ClusterFingerprint
from scipy.spatial import cKDTree as KDTree
from ase.geometry import wrap_positions


class ClusterGenerator(object):
    """
    Class for generating cluster info that is independent of the
    template
    """
    def __init__(self, prim_cell):
        self.with_cutoff = SitesWithinCutoff(self)
        self.prim = prim_cell
        self.shifts = np.zeros((len(prim_cell), 3))
        pos = self.prim.get_positions()
        for i, atom in enumerate(prim_cell):
            self.shifts[atom.tag, :] = pos[i, :]
        self.prim_cell_invT = np.linalg.inv(self.prim.get_cell().T)

    def eucledian_distance_vec(self, x1, x2):
        """
        Eucledian distance between to vectors

        Parameters:
        x1: array int
            First vector
        x2: array int
            Second vector
        """
        cellT = self.prim.get_cell().T
        euc1 = cellT.dot(x1[:3]) + self.shifts[x1[3], :]
        euc2 = cellT.dot(x2[:3]) + self.shifts[x2[3], :]
        return euc2 - euc1

    def cartesian(self, x):
        cellT = self.prim.get_cell().T
        return cellT.dot(x[:3]) + self.shifts[x[3]]

    def get_four_vector(self, pos, lattice):
        """Return the four vector of an atom."""
        if lattice is None:
            lattice = self.get_lattice(pos)

        pos -= self.shifts[lattice]
        int_pos = self.prim_cell_invT.dot(pos)
        four_vec = np.zeros(4, dtype=int)
        four_vec[:3] = np.round(int_pos).astype(int)
        four_vec[3] = lattice
        return four_vec

    def get_lattice(self, pos):
        """
        Return the corresponding sublattice of a cartesian position

        Parameters:
        pos: array of length 3
            Cartesian position
        """
        reshaped = np.reshape(pos, (1, 3))
        wrapped = wrap_positions(reshaped, self.prim.get_cell())
        diff_sq = np.sum((wrapped[0, :] - self.shifts)**2, axis=1)
        return np.argmin(diff_sq)

    def eucledian_distance(self, x1, x2):
        d = self.eucledian_distance_vec(x1, x2)
        return np.sqrt(np.sum(d**2))

    @property
    def shortest_diag(self):
        shortest = None
        cellT = self.prim.get_cell().T
        for w in product([-1, 0, 1], repeat=3):
            if all(x == 0 for x in w):
                continue

            diag = cellT.dot(w)
            length = np.sqrt(np.sum(diag**2))
            if shortest is None or length < shortest:
                shortest = length
        return shortest

    @property
    def num_sub_lattices(self):
        return len(self.prim)

    def sites_within_cutoff(self, cutoff, x0=[0, 0, 0, 0]):
        min_diag = self.shortest_diag
        max_int = int(cutoff/min_diag) + 1

        def filter_func(x):
            d = self.eucledian_distance(x0, x)
            return d > cutoff or d < 1E-5

        sites = filterfalse(
            filter_func,
            product(range(-max_int, max_int+1),
                    range(-max_int, max_int+1),
                    range(-max_int, max_int+1),
                    range(self.num_sub_lattices)))
        return sites

    def get_fp(self, X):
        """
        Generate finger print given a position matrix
        """
        X = np.array(X)
        com = np.mean(X, axis=0)
        X -= com
        X = X.dot(X.T)
        diag = np.diagonal(X)
        off_diag = []
        for i in range(1, X.shape[0]):
            off_diag += np.diagonal(X, offset=i).tolist()
        N = X.shape[0]
        assert len(off_diag) == N*(N-1)/2
        inner = np.array(sorted(diag, reverse=True) + sorted(off_diag))
        fp = ClusterFingerprint(list(inner))
        return fp

    def prepare_within_cutoff(self, cutoff, lattice):
        within_cutoff = {}
        x0 = [0, 0, 0, lattice]
        sites = self.with_cutoff.get(cutoff, lattice)
        sites = set(map(tuple, sites))
        within_cutoff[tuple(x0)] = sites
        for s in sites:
            nearby = [s1 for s1 in sites
                      if self.eucledian_distance(s1, s) <= cutoff and s1 != s]

            within_cutoff[s] = nearby
        return within_cutoff

    def site_iterator(self, within_cutoff, size, ref_lattice):
        """
        Return an iterator of all combinations of sites within a cutoff

        Parameters:

        within_cutoff: dict
            Dictionary returned by `prepare_within_cutoff`

        size: int
            Cluster size

        ref_lattice: int
            Reference lattice
        """
        x0 = (0, 0, 0, ref_lattice)

        def recursive_yield(remaining, current):
            if len(current) == size - 1:
                yield current
            else:
                for next_item in remaining:
                    # Avoid double counting
                    if any(next_item <= v for v in current):
                        continue
                    rem = set.intersection(remaining, within_cutoff[next_item])
                    yield from recursive_yield(rem, current+[next_item])

        for v in within_cutoff[x0]:
            rem = within_cutoff[x0].intersection(within_cutoff[v])
            yield from recursive_yield(rem, [v])

    def generate(self, size, cutoff, ref_lattice=0):
        clusters = []
        all_fps = []
        v0 = [0, 0, 0, ref_lattice]
        x0 = np.array(self.cartesian(v0))
        cutoff_lut = self.prepare_within_cutoff(cutoff, ref_lattice)
        for comb in self.site_iterator(cutoff_lut, size, ref_lattice):
            X = [x0] + [self.cartesian(v) for v in comb]
            fp = self.get_fp(X)
            new_item = [v0] + [list(x) for x in comb]

            # Find the group
            try:
                group = all_fps.index(fp)
                clusters[group].append(new_item)
            except ValueError:
                # Does not exist, create a new group
                clusters.append([new_item])
                all_fps.append(fp)

        # Order the indices by internal indices
        for cluster in clusters:
            for i, f in enumerate(cluster):
                ordered_f = self._order_by_internal_distances(f)
                cluster[i] = ordered_f
        return clusters, all_fps

    def _get_internal_distances(self, figure):
        """
        Return all the internal distances of a figure
        """
        dists = []
        for x0 in figure:
            d = []
            for x1 in figure:
                dist = self.eucledian_distance(x0, x1)
                dist = dist.round(decimals=6)
                d.append(dist.tolist())
            dists.append(sorted(d, reverse=True))
        return dists

    def _order_by_internal_distances(self, figure):
        """Order the indices by internal distances."""
        dists = self._get_internal_distances(figure)
        zipped = sorted(list(zip(dists, figure)), reverse=True)
        return [x[1] for x in zipped]

    def to_atom_index(self, cluster, lut):
        """
        Convert the integer vector representation to an atomic index

        Parameters:
        clusters: list
            List of clusters (in integer vector representation)

        template: Atoms
            Template atoms to use

        lut: dict
            Look up table for 4-vectors to indices
        """
        return [[lut[tuple(ivec)] for ivec in fig] for fig in cluster]

    def equivalent_sites(self, figure):
        """Find the equivalent sites of a figure."""
        dists = self._get_internal_distances(figure)
        equiv_sites = []
        for i in range(len(dists)):
            for j in range(i+1, len(dists)):
                if np.allclose(dists[i], dists[j]):
                    equiv_sites.append((i, j))

        # Merge pairs into groups
        merged = []
        for equiv in equiv_sites:
            found_group = False
            for m in merged:
                if any(x in m for x in equiv):
                    m.update(equiv)
                    found_group = True
            if not found_group:
                merged.append(set(equiv))
        return [list(x) for x in merged]

    def get_max_distance(self, figure):
        """Return the maximum distance of a figure."""
        internal_dists = self._get_internal_distances(figure)
        max_dists = [x[0] for x in internal_dists]
        return max(max_dists)


class SitesWithinCutoff(object):
    """
    Class that keeps track of nearby sites. It automatically re-calculates if
    the cutoff changes. If it can re-use a previous result it will do that.
    """
    def __init__(self, generator):
        self.generator = generator
        self.cutoff = 0.0
        self.pre_calc = {}

    def must_generate(self, cutoff, ref_lattice):
        """
        Check if we can re-use a precalculated result
        """
        if abs(cutoff - self.cutoff) > 1e-10:
            return True
        return ref_lattice not in self.pre_calc.keys()

    def get(self, cutoff, ref_lattice):
        """
        Return sites within the cutoff
        """
        if self.must_generate(cutoff, ref_lattice):
            sites = self.generator.sites_within_cutoff(
                cutoff, [0, 0, 0, ref_lattice])
            self.pre_calc[ref_lattice] = list(sites)
            self.cutoff = cutoff
        return self.pre_calc[ref_lattice]



"""Module for generating new structures for training."""
import os
import numpy as np
from random import shuffle
from copy import deepcopy
from functools import reduce

from ase.io import read
from ase.db import connect
from ase.atoms import Atoms
from ase.utils.structure_comparator import SymmetryEquivalenceCheck

from clease import ClusterExpansionSetting, CorrFunction
from clease.tools import wrap_and_sort_by_position, nested_list2str
from clease.structure_generator import ProbeStructure, GSStructure
from clease import _logger

try:
    from math import gcd
except ImportError:
    from fractions import gcd

max_attempt = 10
max_fail = 10


class MaxAttemptReachedError(Exception):
    """Raised when number of try reaches 10."""

    pass


# class GenerateStructures(object):
class NewStructures(object):
    """Generate new structure in Atoms object format.

    Parameters:

    setting: CEBulk or CESpacegroup object

    generation_number: int
        a generation number to be assigned to the newly generated structure.

    struct_per_gen: int
        number of structures to generate per generation.
    """

    def __init__(self, setting, generation_number=None, struct_per_gen=5):
        if not isinstance(setting, ClusterExpansionSetting):
            msg = "setting must be ClusterExpansionSetting object"
            raise TypeError(msg)
        self.setting = setting
        self.db = connect(setting.db_name)
        self.corrfunc = CorrFunction(self.setting)
        self.struct_per_gen = struct_per_gen

        if generation_number is None:
            self.gen = self._determine_gen_number()
        else:
            self.gen = generation_number

    def num_in_gen(self):
        return len([row.id for row in self.db.select(gen=self.gen)])

    def num_to_gen(self):
        return max(self.struct_per_gen - self.num_in_gen(), 0)

    def generate_probe_structure(self, atoms=None, size=None,
                                 init_temp=None,
                                 final_temp=None, num_temp=5,
                                 num_steps_per_temp=1000,
                                 approx_mean_var=True,
                                 num_samples_var=10000):
        """Generate a probe structure according to PRB 80, 165122 (2009).

        Parameters:

        atoms: Atoms object
            Atoms object with the desired cell size and shape of the new
            structure.

        size: list of length 3
            (ignored if atoms is given)
            If specified, the structure with the provided size is generated.
            If None, the size will be generated randomly with a bias towards
                more cubic cells (i.e., cell with similar magnitudes of vectors
                a, b and c)

        init_temp: int or float
            initial temperature (does not represent *physical* temperature)

        final_temp: int or float
            final temperature (does not represent *physical* temperature)

        num_temp: int
            number of temperatures to be used in simulated annealing

        num_steps_per_temp: int
            number of steps in simulated annealing

        approx_mean_var: bool
            whether or not to use a spherical and isotropical distribution
            approximation scheme for determining the mean variance.
            -'True': Assume a spherical and isotropical distribution of
                     structures in the configurational space.
                     Corresponds to eq.4 in PRB 80, 165122 (2009)
            -'False': Use sigma and mu of eq.3 in PRB 80, 165122 (2009)
                      to characterize the distribution of structures in
                      population.
                      Requires pre-sampling of random structures before
                      generating probe structures.
                      sigma and mu are generated and stored in
                      'probe_structure-sigma_mu.npz' file.

        num_samples_var: int
            Number of samples to be used in determining signam and mu.
            Only used when approx_mean_var is True.

        Note: init_temp and final_temp are automatically generated if either
              one of the two is not specified.
        """

        if not approx_mean_var:
            # check to see if there are files containing mu and sigma values
            if not os.path.isfile('probe_structure-sigma_mu.npz'):
                self._generate_sigma_mu(num_samples_var)

        _logger("Generate {} probe structures (generation: {}, "
                "struct_per_gen={}, {} present)."
                .format(self.num_to_gen(), self.gen, self.struct_per_gen,
                        self.num_in_gen()))

        current_count = 0
        num_attempt = 0
        num_to_generate = self.num_to_gen()

        while current_count < num_to_generate:
            if atoms is not None:
                self.setting.set_active_template(atoms=atoms)
            else:
                self.setting.set_active_template(size=size)
            # Break out of the loop if reached struct_per_gen
            num_struct = self.num_in_gen()
            if num_struct >= self.struct_per_gen:
                break

            struct = self._get_struct_at_conc(conc_type='random')

            _logger('Generating structure {} out of {}.'
                    .format(current_count + 1, num_to_generate))
            ps = ProbeStructure(self.setting, struct, num_to_generate,
                                init_temp, final_temp, num_temp,
                                num_steps_per_temp, approx_mean_var)
            probe_struct, cf = ps.generate()

            # Remove energy from result dictionary
            probe_struct.get_calculator().results.pop('energy')
            formula_unit = self._get_formula_unit(probe_struct)
            if self._exists_in_db(probe_struct, formula_unit):
                msg = 'generated structure is already in DB.\n'
                msg += 'generating again... '
                msg += '{} out of {} attempts'.format(num_attempt + 1,
                                                      max_attempt)
                _logger(msg)
                num_attempt += 1
                if num_attempt >= max_attempt:
                    raise MaxAttemptReachedError("Could not generate probe "
                                                 "structure in {} attempts."
                                                 .format(max_attempt))
            else:
                num_attempt = 0

            _logger('Probe structure generated.\n')
            self.insert_structure(init_struct=probe_struct)
            current_count += 1

    @property
    def corr_func_table_name(self):
        return "{}_cf".format(self.setting.bf_scheme.name)

    def generate_gs_structure_multiple_templates(
            self, num_templates=20, num_prim_cells=2, init_temp=2000,
            final_temp=1, num_temp=10, num_steps_per_temp=1000,
            eci=None):
        """
        Search for ground states in many templates.

        Parameters:

        num_templates: int
            Number of templates to search in. Simmulated annealing is done in
            each cell and the one with the lowest energy is taken as the ground
            state.

        num_prim_cells: int
            Number of primitive cells to use when constructing templates. The
            volume of all the templates used will be
            num_prim_cells*vol_primitive, where vol_primitive is the volume of
            the primitive cell.

        See doc-string of `generate_gs_structure` for the rest of the
        argument.
        """

        for i in range(self.num_to_gen()):
            _logger('Generating ground-state structures: {} of {}'
                    ''.format(i, self.struct_per_gen))

            self._generate_one_gs_structure_multiple_templates(
                num_templates=num_templates, num_prim_cells=num_prim_cells,
                init_temp=init_temp, final_temp=final_temp,
                num_temp=num_temp, num_steps_per_temp=num_steps_per_temp,
                eci=eci
            )

    def _generate_one_gs_structure_multiple_templates(
            self, num_templates=20, num_prim_cells=2, init_temp=2000,
            final_temp=1, num_temp=10, num_steps_per_temp=1000,
            eci=None):
        """
        Search for ground states in many templates.

        Parameters:

        num_templates: int
            Number of templates to search in. Simmulated annealing is done in
            each cell and the one with the lowest energy is taken as the ground
            state.

        num_prim_cells: int
            Number of primitive cells to use when constructing templates. The
            volume of all the templates used will be
            num_prim_cells*vol_primitive, where vol_primitive is the volume of
            the primitive cell.

        See doc-string of `generate_gs_structure` for the rest of the
        argument.
        """
        templates = self.setting.template_atoms.get_fixed_volume_templates(
            num_templates=num_templates, num_prim_cells=num_prim_cells)

        if len(templates) == 0:
            msg = "Could not find any templates with matching the constraints"
            raise RuntimeError(msg)

        self.setting.set_active_template(atoms=templates[0])

        nib = [len(x) for x in self.setting.index_by_basis]
        x = self.setting.concentration.get_random_concentration(nib=nib)
        num_insert = self.setting.concentration.conc_in_int(nib, x)

        energies = []
        gs_structs = []
        for i, atoms in enumerate(templates):
            _logger('Searching for GS in template {} of {}'
                    ''.format(i, len(templates)))
            self.setting.set_active_template(atoms=atoms)

            struct = self._random_struct_at_conc(num_insert)
            es = GSStructure(self.setting, struct, self.struct_per_gen,
                             init_temp, final_temp, num_temp,
                             num_steps_per_temp, eci)

            gs_struct, _ = es.generate()
            gs_structs.append(gs_struct)
            energies.append(gs_struct.get_potential_energy())

        # Find the position of the minimum energy structure
        min_energy_indx = np.argmin(energies)
        gs = gs_structs[min_energy_indx]
        self.setting.set_active_template(atoms=gs)
        self.insert_structure(init_struct=gs_structs[min_energy_indx])

    def generate_gs_structure(self, atoms=None, init_temp=2000,
                              final_temp=1, num_temp=10,
                              num_steps_per_temp=1000, eci=None,
                              random_composition=False):
        """Generate ground-state structure.

        Parameters:

        atoms: Atoms object or a list of Atoms object
            Atoms object with the desired size and composition of the new
            structure. A list of Atoms with different size and/or compositions
            can be passed. Compositions of the supplied Atoms object(s) are
            ignored when random_composition=True.

        init_temp: int or float
            initial temperature (does not represent *physical* temperature)

        final_temp: int or float
            final temperature (does not represent *physical* temperature)

        num_temp: int
            number of temperatures to be used in simulated annealing

        num_steps_per_temp: int
            number of steps in simulated annealing

        eci: dict
            Dictionary containing cluster names and their ECI values

        random_composition: bool
            -*False* and atoms = Atoms object: One ground-state structure with
                matching size and composition of the supplied Atoms object is
                generated
            -*False* and atoms = list: The same number of ground-state
                structures that matches the length of the list is generated
                Note 1: num_struct_per_gen is ignored and all of the generated
                        structures have the same generation number
                Note 2: each GS structure will have matching size and
                        composition of the suplied Atoms objects
            -*True* and atoms = Atoms object: GS structure(s) with a
                matching size of the Atoms object is generated at a random
                composition (within the composition range specified in
                Concentration class)
                Note 1: This will generate GS structures until the number of
                        structures with the current generation number equals
                        num_struct_per_gen
                Note 2: A check is performed to ensure that none of the newly
                        generated GS structures have the same composition
            -*True* and atoms = list: The same number of GS structures that
                matches the length of the list is generated
                Note 1: num_struct_per_gen is ignored and all of the generated
                        structures have the same generation number
                Note 2: each GS structure will have matching sizes of the
                        supplied Atoms objects but with a random composition
                Note 3: No check is performed to ensure that all new GS
                        structures have unique composition
        """
        structs = self._set_initial_structures(atoms, random_composition)
        current_count = 0
        num_attempt = 0
        num_to_generate = min([self.num_to_gen(), len(structs)])
        while current_count < num_to_generate:
            struct = structs[current_count].copy()
            self.setting.set_active_template(atoms=struct)
            _logger("Generating structure {} out of {}."
                    .format(current_count + 1, num_to_generate))
            es = GSStructure(self.setting, struct, self.struct_per_gen,
                             init_temp, final_temp, num_temp,
                             num_steps_per_temp, eci)
            gs_struct, cf = es.generate()
            formula_unit = self._get_formula_unit(gs_struct)

            if self._exists_in_db(gs_struct, formula_unit):
                msg = 'generated structure is already in DB.\n'
                msg += 'generating again... '
                msg += '{} out of {} attempts'.format(num_attempt + 1,
                                                      max_attempt)
                _logger(msg)
                num_attempt += 1
                if num_attempt >= max_attempt:
                    raise MaxAttemptReachedError(
                        "Could not generate ground-state structure in "
                        "{} attempts.".format(max_attempt))
                continue
            else:
                num_attempt = 0

            min_energy = gs_struct.get_potential_energy()
            msg = 'Structure with E = {:.3f} generated.\n'.format(min_energy)
            _logger(msg)
            self.insert_structure(init_struct=gs_struct)
            current_count += 1

    def generate_random_structures(self, atoms=None):
        """Generate a given number of random structures.

        Parameters:

        atoms: Atoms object or None.
            If Atoms object is passed, the passed object will be used as a
            template for all the random structures being generated.
            If None, a random template will be chosen
            (different for each structure)
        """
        _logger("Generating {} random structures "
                "(generation: {}, struct_per_gen={}, {} present)"
                .format(self.num_to_gen(), self.gen, self.struct_per_gen,
                        self.num_in_gen()))

        fail_counter = 0
        i = 0

        num_structs = self.num_to_gen()
        while i < num_structs and fail_counter < max_fail:
            if self.generate_one_random_structure(atoms=atoms, verbose=False):
                _logger("Generated {} random structures".format(i + 1))
                i += 1
                fail_counter = 0
            else:
                fail_counter += 1

        if fail_counter >= max_fail:
            RuntimeError("Could not find a structure that does not exist in "
                         "DB after {} attempts."
                         "".format(int(max_attempt * max_fail)))

    def generate_one_random_structure(self, atoms=None, verbose=True):
        """Generate a random structure.

        Inserts a new structure to database if a unique structure is found.

        Returns ``True`` if unique structure is found and inserted in DB,
        ``False`` otherwise.

        Parameters:

        atoms: Atoms object or None.
            If Atoms object is passed, the passed object will be used as a
            template for all the random structures being generated.
            If None, a random template will be chosen
            (different for each structure)

        verbose: bool
            If ``True``, print error message when it reaches max. number of
            attempts.
        """
        num_attempts = 0
        unique_structure_found = False

        while not unique_structure_found and num_attempts < max_attempt:
            self.setting.set_active_template(atoms=atoms)
            new_atoms = self._get_struct_at_conc(conc_type="random")
            fu = self._get_formula_unit(new_atoms)
            if not self._exists_in_db(new_atoms, fu):
                unique_structure_found = True
            else:
                num_attempts += 1

        if not unique_structure_found:
            if verbose:
                _logger("Could not find a structure that does not already "
                        "exist in the DB within {} attempts."
                        "".format(max_attempt))
            return False

        self.insert_structure(init_struct=new_atoms)
        return True

    def _set_initial_structures(self, atoms, random_composition=False):
        structs = []
        if isinstance(atoms, Atoms):
            struct = wrap_and_sort_by_position(atoms)
            if random_composition is False:
                num_to_gen = 1
                _logger("Generate 1 ground-state structure.")
                structs.append(struct)
            else:
                _logger("Generate {} ground-state structures "
                        "(generation: {}, struct_per_gen={}, {} present)"
                        .format(self.num_to_gen(),
                                self.gen,
                                self.struct_per_gen,
                                self.num_in_gen()))
                self.setting.set_active_template(atoms=struct)
                num_to_gen = self.num_to_gen()
                concs = []
                # Get unique concentrations
                num_attempt = 0
                nib = [len(x) for x in self.setting.index_by_basis]
                while len(concs) < num_to_gen:
                    x = self.setting.concentration.get_random_concentration(
                        nib=nib)
                    if True in [np.allclose(x, i) for i in concs]:
                        num_attempt += 1
                    else:
                        concs.append(x)
                        num_attempt = 0

                    if num_attempt > 100:
                        raise RuntimeError("Could not find {} unique "
                                           "compositions using the provided "
                                           "Atoms object"
                                           .format(self.num_to_gen))
                num_atoms_in_basis = [len(indices) for indices
                                      in self.setting.index_by_basis]
                for x in concs:
                    num_insert = self.setting.concentration.conc_in_int(
                        num_atoms_in_basis, x)
                    structs.append(self._random_struct_at_conc(num_insert))

        elif all(isinstance(a, Atoms) for a in atoms):
            _logger("Generate {} ground-state structures ".format(len(atoms)))
            if random_composition is False:
                for struct in atoms:
                    structs.append(wrap_and_sort_by_position(struct))
            else:
                concs = []
                nib = [len(x) for x in self.setting.index_by_basis]
                for struct in atoms:
                    self.setting.set_active_template(atoms=struct)
                    x = self.setting.concentration.get_random_concentration(
                        nib=nib)
                    num_atoms_in_basis = [len(indices) for indices
                                          in self.setting.index_by_basis]
                    num_insert = self.setting.concentration.conc_in_int(
                        num_atoms_in_basis, x)
                    structs.append(self._random_struct_at_conc(num_insert))

        else:
            raise ValueError("atoms must be either an Atoms object or a list "
                             "of Atoms objects")
        return structs

    def generate_initial_pool(self, atoms=None):
        """Generate initial pool of structures.

        Initial pool of structures are generated, in sequence, using

        1. generate_conc_extrema(): structures at concentration where the
           number of consituting elements is at its max/min.
        2. generate_random_structures(): random structures are random
           concentration.

        Parameters:

        atoms: Atoms object | None.
            If Atoms object is passed, the passed object will be used as a
            template for all the random structures being generated.
            If None, a random template will be chosen
            (different for each structure)."""

        self.generate_conc_extrema()
        self.generate_random_structures(atoms=atoms)

    def generate_conc_extrema(self):
        """Generate initial pool of structures with max/min concentration."""
        from itertools import product
        _logger("Generating one structure per concentration where the number "
                "of an element is at max/min")
        indx_in_each_basis = []
        start = 0
        for basis in self.setting.concentration.basis_elements:
            indx_in_each_basis.append(list(range(start, start + len(basis))))
            start += len(basis)

        for indx in product(*indx_in_each_basis):
            atoms = self._get_struct_at_conc(conc_type="max", index=indx)
            atoms = wrap_and_sort_by_position(atoms)
            self.insert_structure(init_struct=atoms)

    def _get_struct_at_conc(self, conc_type='random', index=0):
        """Generate a structure at a concentration specified.

        Parameters:

        conc_type: str
            One of 'min', 'max' and 'random'

        index: int
            index of the flattened basis_element array to specify which
            element to be maximized/minimized
        """
        conc = self.setting.concentration
        if conc_type == 'min':
            x = conc.get_conc_min_component(index)
        elif conc_type == 'max':
            x = conc.get_conc_max_component(index)
        else:
            nib = [len(x) for x in self.setting.index_by_basis]
            x = conc.get_random_concentration(nib=nib)

        num_atoms_in_basis = [len(indices) for indices
                              in self.setting.index_by_basis]
        num_atoms_to_insert = conc.conc_in_int(num_atoms_in_basis, x)
        atoms = self._random_struct_at_conc(num_atoms_to_insert)

        return atoms

    def insert_structures(self, traj_init=None, traj_final=None,
                          cb=lambda num, tot: None):
        """
        Insert a sequence of initial and final structures from trajectory.

        Parameters:

        traj_init: str
            Filename of a trajectory file with initial structures

        traj_final: str
            Filename of a trajectory file with the final structures

        cb: function
            Callback function that is called every time a structure is inserted
            (or rejected because it exists before). The signature of the
            function is cb(num, tot) where num is the number of inserted
            structure and tot is the total number of structures that should
            be inserted
        """
        from ase.io.trajectory import TrajectoryReader
        from clease.tools import count_atoms
        traj_in = TrajectoryReader(traj_init)

        if traj_final is None:
            for i, init in enumerate(traj_in):
                self.insert_structure(init_struct=init)
                cb(i+1, len(traj_in))
            return

        traj_final = TrajectoryReader(traj_final)

        if len(traj_in) != len(traj_final):
            raise ValueError("Different number of structures in "
                             "initial trajectory file and final.")

        num_ins = 0
        for init, final in zip(traj_in, traj_final):
            # Check that composition (except vacancies matches)
            count_init = count_atoms(init)
            count_final = count_atoms(final)
            for k in count_final.keys():
                if k not in count_init.keys():
                    raise ValueError("Final and initial structure contains "
                                     "different elements")

                if count_init[k] != count_final[k]:
                    raise ValueError("Final and initial structure has "
                                     "different number of each species")

            self.insert_structure(init_struct=init, final_struct=final)
            num_ins += 1
            cb(num_ins, len(traj_in))

    def insert_structure(self, init_struct=None, final_struct=None, name=None):
        """Insert a user-supplied structure to the database.

        Parameters:

        init_struct: .xyz, .cif or .traj file
            Unrelaxed initial structure.

        final_struct: .traj file (optional)
            Final structure that contains the energy.
            Needs to also supply init_struct in order to use the final_struct.

        name: str (optional)
            Name of the DB entry if a custom name is to be used.
            If ``None``, default naming convention will be used.
        """
        if init_struct is None:
            raise TypeError('init_struct must be provided')

        if name is not None:
            num = sum(1 for _ in self.db.select(['name', '=', name]))
            if num > 0:
                raise ValueError("Name: {} already exists in DB!".format(name))

        if isinstance(init_struct, Atoms):
            init = wrap_and_sort_by_position(init_struct)
        else:
            init = wrap_and_sort_by_position(read(init_struct))

        self.setting.set_active_template(atoms=init_struct)

        formula_unit = self._get_formula_unit(init)
        if self._exists_in_db(init, formula_unit):
            _logger('Supplied structure already exists in DB.'
                    'The structure will not be inserted.')
            return

        cf = self.corrfunc.get_cf(init)
        kvp = self._get_kvp(init, formula_unit)

        if name is not None:
            kvp['name'] = name

        kvp['converged'] = False
        kvp['started'] = False
        kvp['queued'] = False
        kvp['struct_type'] = 'initial'
        tab_name = self.corr_func_table_name
        uid_init = self.db.write(init, kvp, external_tables={tab_name: cf})

        if final_struct is not None:
            if isinstance(final_struct, Atoms):
                final = final_struct
            else:
                final = read(final_struct)
            kvp_final = {'struct_type': 'final', 'name': kvp['name']}
            uid = self.db.write(final, kvp_final)
            self.db.update(uid_init, converged=True, started='', queued='',
                           final_struct_id=uid)

    def _exists_in_db(self, atoms, formula_unit=None):
        """Check to see if the passed atoms already exists in DB.

        To reduce the number of assessments for symmetry equivalence,
        check is only performed with the entries with the same concentration
        value.

        Return *True* if there is a symmetry-equivalent structure in DB,
        return *False* otherwise.

        Parameters:

        atoms: Atoms object

        formula_unit: str
            reduced formula unit of the passed Atoms object
        """
        cond = [('name', '!=', 'template'), ('name', '!=', 'primitive_cell')]
        if formula_unit is not None:
            cond.append(("formula_unit", "=", formula_unit))

        to_prim = True
        try:
            __import__('spglib')
        except Exception:
            msg = "Warning! Setting to_primitive=False because spglib "
            msg += "is missing!"
            _logger(msg)
            to_prim = False

        symmcheck = SymmetryEquivalenceCheck(angle_tol=1.0, ltol=0.05,
                                             stol=0.05, scale_volume=True,
                                             to_primitive=to_prim)
        atoms_in_db = []
        for row in self.db.select(cond):
            atoms_in_db.append(row.toatoms())
        return symmcheck.compare(atoms.copy(), atoms_in_db)

    def _get_kvp(self, atoms, formula_unit=None):
        """Get key-value pairs of the passed Atoms object.

        Create key-value pairs and return it.

        Parameters:

        atoms: Atoms object

        formula_unit: str
            reduced formula unit of the passed Atoms object
        """
        if formula_unit is None:
            raise ValueError("Formula unit not specified!")
        kvp = {}
        kvp['gen'] = self.gen
        kvp['converged'] = False
        kvp['started'] = False
        kvp['queued'] = False

        count = 0
        for _ in self.db.select(formula_unit=formula_unit):
            count += 1
        kvp['name'] = formula_unit + "_{}".format(count)
        kvp['formula_unit'] = formula_unit
        kvp['struct_type'] = 'initial'
        kvp['size'] = nested_list2str(self.setting.size)
        return kvp

    def _get_formula_unit(self, atoms):
        """Generates a reduced formula unit for the structure."""
        atom_count = []
        all_nums = []
        for group in self.setting.index_by_basis:
            new_count = {}
            for indx in group:
                symbol = atoms[indx].symbol
                if symbol not in new_count.keys():
                    new_count[symbol] = 1
                else:
                    new_count[symbol] += 1
            atom_count.append(new_count)
            all_nums += [v for k, v in new_count.items()]
        gcdp = reduce(lambda x, y: gcd(x, y), all_nums)
        fu = ""
        for i, count in enumerate(atom_count):
            keys = list(count.keys())
            keys.sort()
            for k in keys:
                fu += "{}{}".format(k, int(count[k] / gcdp))
            if i < len(atom_count) - 1:
                fu += "_"
        return fu

    def _random_struct_at_conc(self, num_atoms_to_insert):
        """Generate a random structure."""
        rnd_indices = []
        for indices in self.setting.index_by_basis:
            rnd_indices.append(deepcopy(indices))
            shuffle(rnd_indices[-1])

        # Insert the number of atoms
        basis_elem = self.setting.concentration.basis_elements
        assert len(rnd_indices) == len(basis_elem)
        atoms = self.setting.atoms.copy()
        current_conc = 0
        num_atoms_inserted = 0
        for basis in range(len(rnd_indices)):
            current_indx = 0
            for symb in basis_elem[basis]:
                for _ in range(num_atoms_to_insert[current_conc]):
                    atoms[rnd_indices[basis][current_indx]].symbol = symb
                    current_indx += 1
                    num_atoms_inserted += 1
                current_conc += 1
        assert num_atoms_inserted == len(atoms)
        return atoms

    def _determine_gen_number(self):
        """Determine generation number based on the values in DB."""
        try:
            gens = [row.get('gen') for row in self.db.select()]
            gens = [i for i in gens if i is not None]
            gen = max(gens)
            num_in_gen = len([row.id for row in self.db.select(gen=gen)])
            if num_in_gen >= self.struct_per_gen:
                gen += 1
        except ValueError:
            gen = 0
        return gen

    def _generate_sigma_mu(self, num_samples_var):
        _logger('===========================================================\n'
                'Determining sigma and mu value for assessing mean variance.\n'
                'May take a long time depending on the number of samples \n'
                'specified in the *num_samples_var* argument.\n'
                '===========================================================')
        count = 0
        cfm = np.zeros((num_samples_var, len(self.setting.all_cf_names)),
                       dtype=float)
        while count < num_samples_var:
            atoms = self._get_struct_at_conc(conc_type='random')
            cfm[count] = self.corrfunc.get_cf(atoms, 'array')
            count += 1
            _logger('sampling {} ouf of {}'.format(count, num_samples_var))

        sigma = np.cov(cfm.T)
        mu = np.mean(cfm, axis=0)
        np.savez('probe_structure-sigma_mu.npz', sigma=sigma, mu=mu)



import json


class ClusterFingerprint(object):
    def __init__(self, fp, tol=1E-9):
        self.fp = fp
        self.tol = tol

    def __lt__(self, other):
        if len(self.fp) < len(other.fp):
            return True
        elif len(self.fp) > len(other.fp):
            return False

        for x, y in zip(self.fp, other.fp):
            diff = x - y
            if diff < -self.tol:
                return True
            elif diff > self.tol:
                return False
        return False

    def __eq__(self, other):
        if len(self.fp) != len(other.fp):
            return False

        for x, y in zip(self.fp, other.fp):
            if abs(x - y) > self.tol:
                return False
        return True

    def __getitem__(self, i):
        return self.fp[i]

    def __len__(self):
        return len(self.fp)

    def __str__(self):
        return json.dumps(self.todict())

    def todict(self):
        return {
            'tol': self.tol,
            'fp': self.fp
        }

    def fromJSON(self, data):
        self.tol = data['tol']
        self.fp = list(data['fp'])

    @staticmethod
    def load(data):
        fp = ClusterFingerprint(None)
        fp.fromJSON(data)
        return fp



import time
from clease import _logger
from clease.montecarlo.constraints import CollectiveVariableConstraint
import json
import numpy as np
from ase.units import kB
from copy import deepcopy


class NoneNotAcceptedError(Exception):
    pass


class PeakNotAcceptedError(Exception):
    pass


class MetaDynamicsSampler(object):
    """
    Class for performing meta dynamics sampler

    Parameters:

    mc: Montecarlo
        A Monte Carlo sampler

    bias: BiasPotential
        A bias potential that should be altered in order to recover the free
        energy

    flat_limit: float
        The histogram of visits is considered flat, when the minimum value
        is larger than flat_limit*np.mean(hist)

    mod_factor: float
        Modification factor in units of kB*T

    fname: str
        Filename used to store the simulation state when finished
    """
    def __init__(self, mc=None, bias=None, flat_limit=0.8,
                 mod_factor=0.1, fname='metadyn.json'):
        self.mc = mc
        self.bias = bias
        self.mc.add_bias(self.bias)
        cnst = CollectiveVariableConstraint(
            xmin=self.bias.xmin, xmax=self.bias.xmax, getter=self.bias.getter)
        self.mc.add_constraint(cnst)
        self.mc.update_current_energy()
        self.mc.attach(self.bias.getter, interval=1)
        self.visit_hist = deepcopy(bias)
        self.visit_hist.zero()
        self.flat_limit = flat_limit
        self.mod_factor = mod_factor*kB*self.mc.T
        self.log_freq = 30
        self.fname = fname
        self.progress_info = {'mean': 0.0, 'minimum': 0.0}
        self.observers = []
        self.quit = False

    def _getter_accepts_none(self):
        """Return True if the getter accepts None."""
        try:
            x = self.bias.getter(None)
            x = float(x)
        except Exception:
            return False
        return True

    def _getter_accepts_peak(self):
        """Return True if the getter supports the peak keyword."""
        try:
            x = self.bias.getter([], peak=True)
            x = float(x)
        except Exception:
            return False
        return True

    def add_observer(self, obs, interval=1):
        """Add observer.

        Parameters:

        obs: callable object
            A callable object that takes no arguments

        interval: int
            Will be called at even intervals given by this number
        """
        self.observers.append((obs, interval))

    def visit_is_flat(self):
        """Return True if the histogram of visits is flat."""
        i_min = self.visit_hist.get_index(self.visit_hist.xmin)
        i_max = self.visit_hist.get_index(self.visit_hist.xmax)
        coeff = self.visit_hist.get_coeff()[i_min:i_max]
        avg = np.mean(coeff)

        # Use min and not np.min. It looks like np.min behaves weird
        # when it is running on a worker thread
        minval = min(coeff.tolist())
        self.progress_info['mean'] = avg

        if avg > 0.0:
            self.progress_info['minval'] = minval/avg
        else:
            self.progress_info['minval'] = 0.0

        if np.max(avg) == 0:
            return False
        return minval > self.flat_limit*avg

    def update(self):
        """Update bias potential and visit histogram."""
        x = self.bias.getter(None)
        cur_value = self.bias.evaluate(x)
        self.bias.local_update(x, self.mod_factor)

        new_value = self.bias.evaluate(x)
        self.mc.current_energy += (new_value - cur_value)
        self.visit_hist.local_update(x, 1)

    def run(self, max_sweeps=None):
        """
        Run the calculation.

        Parameters:

        max_sweeps: int or None
            If given, the simulation terminates when this number of sweeps
            is reached
        """
        if not self._getter_accepts_none():
            raise NoneNotAcceptedError("Observer does not accept None as a "
                                       "system change")

        if not self._getter_accepts_peak():
            raise PeakNotAcceptedError("Observer does not accept peak as a "
                                       "keyword argument to __call__")

        if not hasattr(self.bias, 'get_coeff'):
            raise ValueError('The bias potential needs to have a method ',
                             'called get_coeff(), which returns a histogram '
                             'representation of the bias potential')

        if not hasattr(self.bias, 'local_update'):
            raise ValueError('The bias potential needs to have a method '
                             'called local_update(x, dE) which allows a local '
                             'update at position x')
        conv = False
        now = time.time()

        sweep_no = 0
        counter = 0
        _logger('Starting metadynamics sampling...')
        _logger('Writing result to {} every {} sec'.format(self.fname,
                                                           self.log_freq))
        while not conv:
            counter += 1
            if time.time() - now > self.log_freq:
                msg = 'Sweep no. {} '.format(int(counter/len(self.mc.atoms)))
                msg += 'Average visits: {:.2e}. Min/avg: {:.2e}'.format(
                    self.progress_info['mean'],
                    self.progress_info['minval'])
                msg += ' x: {:.2e}'.format(self.bias.getter(None))
                _logger(msg)
                self.save()
                now = time.time()

            self.mc._mc_step()
            self.update()
            if self.visit_is_flat():
                conv = True

            sweep_no = int(counter/len(self.mc.atoms))

            if max_sweeps is not None:
                if sweep_no > max_sweeps:
                    _logger('Reached max number of sweeps...')
                    conv = True

            for obs, interval in self.observers:
                if counter % (interval*len(self.mc.atoms)) == 0:
                    obs()

            if self.quit:
                break

        _logger('Results from metadynamics sampling written to '
                '{}'.format(self.fname))
        self.save()

    def save(self):
        """Save the free energy result to a file."""
        pot = self.bias.todict()
        xmin = self.bias.xmin
        xmax = self.bias.xmax
        x = np.linspace(xmin, xmax, 200)
        beta = 1.0/(kB*self.mc.T)
        betaG = [self.bias.evaluate(y)*beta for y in x]
        data = {
            'bias_pot': pot,
            'betaG': {
                'x': x.tolist(),
                'y': betaG
            }
        }
        with open(self.fname, 'w') as out:
            json.dump(data, out)



"""Class containing a manager for creating template atoms."""
import numpy as np
from itertools import product
from random import shuffle
from ase.build import make_supercell
from clease import SkewnessFilter, EquivalentCellsFilter
from clease.template_filters import CellFilter, AtomsFilter
from clease.tools import all_integer_transform_matrices


class TemplateAtoms(object):
    def __init__(self, prim_cell, supercell_factor=None, size=None,
                 skew_threshold=4, filters=[]):
        if size is None and supercell_factor is None:
            raise TypeError("Either size or supercell_factor needs to be "
                            "specified.\n size: list or numpy array.\n "
                            "supercell_factor: int")

        self.supercell_factor = supercell_factor
        self.size = size

        self.cell_filters = []
        self.atoms_filters = []

        self.add_cell_filter(SkewnessFilter(skew_threshold))
        self.all_cells = []
        self.add_cell_filter(EquivalentCellsFilter(self.all_cells))

        for f in filters:
            if isinstance(f, CellFilter):
                self.add_cell_filter(f)
            elif isinstance(f, AtomsFilter):
                self.add_atoms_filter(f)

        if self.size is not None:
            check_valid_conversion_matrix(self.size)
        self.skew_threshold = skew_threshold
        self.prim_cell = prim_cell

    def __str__(self):
        """Print a summary of the class."""
        msg = "=== TemplateAtoms ===\n"
        msg += "Supercell factor: {}\n".format(self.supercell_factor)
        msg += "Skewed threshold: {}\n".format(self.skew_threshold)
        msg += "Template sizes:\n"
        for size in self.templates['size']:
            msg += "{}\n".format(size)
        return msg

    def add_cell_filter(self, cell_filter):
        """Attach a new Cell filter."""
        if not isinstance(cell_filter, CellFilter):
            raise TypeError("filter has to be an instance of CellFilter")
        self.cell_filters.append(cell_filter)

    def add_atoms_filter(self, at_filter):
        """Attach a new Atoms filter."""
        if not isinstance(at_filter, AtomsFilter):
            raise TypeError("filter has to be an instance of AtomsFilter")
        self.atoms_filters.append(at_filter)

    def clear_filters(self):
        """Remove all filters."""
        self.cell_filters = []
        self.atoms_filters = []

    def remove_filter(self, f):
        """Remove one filter"""
        if isinstance(f, AtomsFilter):
            self.atoms_filters.remove(f)
        elif isinstance(f, CellFilter):
            self.cell_filters.remove(f)
        else:
            raise TypeError('Only AtomsFilters and CellFilters '
                            'can be removed')

    def remove_filters(self, filters):
        """Remove a list of filters."""
        for f in filters:
            self.remove_filter(f)

    def is_valid(self, atoms=None, cell=None):
        """
        Check the validity of the template.

        Return `True` if templates are valid according to the attached filters.

        Parameters:

        atoms: Atoms object

        cell: unit cell vector
        """
        if atoms is None and cell is None:
            msg = "At least one of `atoms` or `cell` must be specified."
            raise ValueError(msg)

        cell_valid = True
        if cell is not None:
            cell_valid = all([f(cell) for f in self.cell_filters])
        elif atoms is not None:
            cell = atoms.get_cell()
            cell_valid = all([f(cell) for f in self.cell_filters])

        if not cell_valid:
            return False

        atoms_valid = True
        if atoms is not None:
            atoms_valid = all([f(atoms) for f in self.atoms_filters])

        return cell_valid and atoms_valid

    @property
    def largest_template_by_num_atom(self):
        """Return the largest template based on number of atoms it has."""
        max_num = 0
        largest_template = None
        for atoms in self.templates['atoms']:
            if len(atoms) > max_num:
                largest_template = atoms
                max_num = len(atoms)
        return largest_template

    @property
    def largest_template_by_diag(self):
        """Return the largest template based on the shortest diagonal."""
        length = 0.0
        largest_template = None
        for atoms in self.templates['atoms']:
            diag_lengths = []
            cell = atoms.get_cell().T
            for w in product([-1, 0, 1], repeat=3):
                if np.allclose(w, 0):
                    continue
                diag = cell.dot(w)
                length = np.sqrt(diag.dot(diag))
                diag_lengths.append(length)

            min_length = np.min(diag_lengths)

            if min_length > length:
                largest_template = atoms
                length = min_length

        return largest_template

    def get_template_with_given_size(self, size):
        """Get the UID of the template with given size.

        Parameters:

        size: list of length 3
        """
        if not is_3x3_matrix(size):
            raise ValueError("Expect the size to be a list of 3x3 matrix")

        check_valid_conversion_matrix(size)

        template = make_supercell(self.prim_cell, size)

        if isinstance(size, np.ndarray):
            size = size.tolist()
        template.info['size'] = size

        if not self.is_valid(atoms=template):
            raise ValueError("Requested size violates the constraints!")
        return template

    def get_template_matching_atoms(self, atoms):
        """Get the UID for the template matching atoms.

        Parameters:

        atoms: Atoms object
            structure to compare its size against template atoms
        """

        size = self._get_conversion_matrix(atoms)
        assert is_3x3_matrix(size)

        prim_cell = self.prim_cell

        if isinstance(size, np.ndarray):
            size = size.tolist()

        template = make_supercell(prim_cell, size)
        template.info['size'] = size

        if not np.allclose(atoms.get_cell(), template.get_cell()):
            raise ValueError("Inconsistent cells! Passed atoms\n"
                             "{}\nGenerated template\n"
                             "".format(atoms.get_cell(), template.get_cell()))

        if not self.is_valid(atoms=template):
            raise ValueError("Requested template violates the constraints!")
        return template

    def _get_conversion_matrix(self, atoms):
        """Return the conversion matrix factor."""
        prim_cell = self.prim_cell

        small_cell = prim_cell.get_cell()
        inv_cell = np.linalg.inv(small_cell)

        large_cell = atoms.get_cell()
        size_factor = large_cell.dot(inv_cell)
        scale_int = size_factor.round(decimals=0).astype(int)
        if np.allclose(size_factor, scale_int):
            check_valid_conversion_matrix(scale_int)
            return scale_int.tolist()

        raise ValueError("The passed atoms object cannot be described by "
                         "repeating of the unit cells. Scale factors found "
                         "{}".format(size_factor))

    def get_all_templates(self):
        """
        Return a list with all templates
        """
        if self.size is not None:
            return [self.get_template_with_given_size(self.size)]

        cells = []
        templates = []
        equiv_filter = EquivalentCellsFilter(cells)
        self.add_cell_filter(equiv_filter)
        ucell = self.prim_cell.get_cell()
        for vol in range(1, self.supercell_factor):
            matrices = all_integer_transform_matrices(vol)
            for mat in matrices:
                cell = mat.dot(ucell)
                if self.is_valid(cell=cell):
                    at = make_supercell(self.prim_cell, mat)
                    if self.is_valid(atoms=at):
                        at.info['size'] = mat.tolist()
                        cells.append(cell)
                        templates.append(at)
        self.remove_filter(equiv_filter)
        return templates

    def get_all_scaled_templates(self):
        """
        Return all templates that can be obtained by scaling the primitive cell
        """
        if self.size is not None:
            return [self.get_template_with_given_size(self.size)]
        cells = []
        templates = []
        equiv_filter = EquivalentCellsFilter(cells)
        self.add_cell_filter(equiv_filter)
        ucell = self.prim_cell.get_cell()
        for diag in product(range(1, self.supercell_factor+1), repeat=3):
            if np.prod(diag) > self.supercell_factor:
                continue

            mat = np.diag(diag)
            cell = mat.dot(ucell)
            if self.is_valid(cell=cell):
                at = make_supercell(self.prim_cell, mat)
                if self.is_valid(atoms=at):
                    cells.append(cell)
                    at.info['size'] = mat.tolist()
                    templates.append(at)
        self.remove_filter(equiv_filter)
        return templates

    def weighted_random_template(self):
        """Select a random template atoms with a bias towards a cubic cell.

        The bias is towards cells that have similar values for x-, y- and
        z-dimension sizes.
        """
        if self.size is not None:
            return self.get_template_with_given_size(self.size)

        ucell = self.prim_cell.get_cell()
        max_attempts = 100000
        exp_value = self.supercell_factor**(1.0/3.0)
        for _ in range(max_attempts):
            diag = np.random.poisson(lam=exp_value, size=3)
            while np.prod(diag) > self.supercell_factor or np.prod(diag) == 0:
                diag = np.random.poisson(lam=exp_value, size=3)

            off_diags = np.random.poisson(lam=1, size=3)

            # Check that off diagonals are smaller or equal to diagonal
            if (off_diags[0] > diag[0] or off_diags[1] > diag[0] or
               off_diags[2] > diag[1]):
                continue

            matrix = np.zeros((3, 3), dtype=int)
            matrix[0, 0] = diag[0]
            matrix[1, 1] = diag[1]
            matrix[2, 2] = diag[2]
            matrix[0, 1] = off_diags[0]
            matrix[0, 2] = off_diags[1]
            matrix[1, 2] = off_diags[2]
            cell = matrix.dot(ucell)

            if self.is_valid(cell=cell):
                atoms = make_supercell(self.prim_cell, matrix)
                if self.is_valid(atoms=atoms):
                    atoms.info['size'] = matrix.tolist()
                    return atoms

        raise RuntimeError("Did not manage to generate a random template that "
                           "satisfies all the constraints")

    def has_atoms_filters(self):
        return len(self.atoms_filters) > 0

    def get_fixed_volume_templates(self, num_prim_cells=10, num_templates=10):
        # Set up a filter that listens to the templates with fixed volume
        from ase.build.tools import niggli_reduce_cell
        cells = []
        transform_matrices = []

        equiv_filter = EquivalentCellsFilter(cells)
        self.add_cell_filter(equiv_filter)

        ucell = self.prim_cell.get_cell()
        matrices = list(all_integer_transform_matrices(num_prim_cells))
        shuffle(matrices)
        for mat in matrices:
            sc = mat.dot(ucell)
            sc, _ = niggli_reduce_cell(sc)
            if self.is_valid(cell=sc):
                # If Atoms filters are present we check if it is valid
                at_valid = True
                if self.has_atoms_filters():
                    atoms = make_supercell(self.prim_cell, mat)
                    at_valid = self.is_valid(atoms=atoms)
                if at_valid:
                    cells.append(sc)
                    transform_matrices.append(mat)

            if len(transform_matrices) >= num_templates:
                break

        templates = []

        if len(transform_matrices) > num_templates:
            shuffle(transform_matrices)
            transform_matrices = transform_matrices[:num_templates]

        for P in transform_matrices:
            atoms = make_supercell(self.prim_cell, P)
            templates.append(atoms)

        # Remove the filter that was artificially added
        self.cell_filters.remove(equiv_filter)
        return templates


def is_3x3_matrix(array):
    return np.array(array).shape == (3, 3)


def check_valid_conversion_matrix(array):
    """
    Make sure that we have a right-handed coordinate system.
    Raise a ValueError if the matrix is not valid.
    """
    determinant = np.linalg.det(array)
    if determinant < 0.0:
        raise ValueError("The determinant of the size matrix is less than "
                         "zero (got {}). For a right "
                         "coordinate system, we need a positive "
                         "determinant!".format(determinant))



from clease.montecarlo.observers import MCObserver
import numpy as np
import time


class EnergyPlotUpdater(MCObserver):
    def __init__(self, energy_obs=None, graph=None, mean_plot=None):
        self.energy_obs = energy_obs
        self.graph = graph
        self.mean_plot = mean_plot

    def __call__(self, system_changes):
        e = self.energy_obs.mean_energies
        e_inst = self.energy_obs.energies
        xmax = len(e) + 1
        ymin = np.min(e) - e[0]
        ymax = np.max(e) - e[0]

        if (abs(ymax - ymin) < 1E-6):
            return

        self.mean_plot.points = [(i, x - e[0]) for i, x in enumerate(e)]
        y_rng = ymax - ymin
        ymin -= 0.05*y_rng
        ymax += 0.05*y_rng
        self.graph.xmax = int(xmax)
        self.graph.ymin = float(ymin)
        self.graph.ymax = float(ymax)
        self.graph.y_ticks_major = float(ymax - ymin)/10.0
        self.graph.x_ticks_major = float(xmax)/10.0
        time.sleep(0.01)



import numpy as np
from math import gcd
from scipy.spatial import cKDTree as KDTree
from ase.geometry import wrap_positions
from ase.build import make_supercell
from clease.tools import wrap_and_sort_by_position
from ase.io import write


class AtomsManager(object):
    """
    Manager class for the Atoms object used in a cluster expansion context.
    This class can return indices of the Atoms object grouped according to
    various schemes.
    """

    def __init__(self, atoms):
        self.atoms = atoms

    def index_by_tag(self):
        """Return atomic indices that are grouped by their tags.

        This method assumes that all atoms are tagged and the tags are in a
        continuous sequence starting from 0.
        """
        tags = set()
        for atom in self.atoms:
            tags.add(atom.tag)

        tags = sorted(list(tags))
        ind_by_tag = [[] for _ in tags]

        for atom in self.atoms:
            ind_by_tag[atom.tag].append(atom.index)
        return ind_by_tag

    def index_by_symbol(self, symbols):
        """Group atomic indices by its atomic symbols.


        Example:

            If symbols = ['Au', 'Cu'], then indices where the indices of Au are
            returned as group 1 (first list in the nested list of indices) and
            the indices of Cu are returned as group 2 (second list in the
            nested list of indices).

            If symbols = ['Au, ['Cu', 'X'], 'Ag'], then indices of Au are
            returned as group 1, the indices of Cu and X are returned as group
            2 and the indices of Ag are returned as group 3.

        Parameters:

        symbols: list
            List with symbols that define a group
        """
        ind_by_symbol = [[] for _ in symbols]
        group_map = {}
        for i, item in enumerate(symbols):
            if isinstance(item, list):
                for x in item:
                    group_map[x] = i
            else:
                group_map[item] = i

        # Loop over indices
        for atom in self.atoms:
            ind_by_symbol[group_map[atom.symbol]].append(atom.index)
        return ind_by_symbol

    def unique_elements(self, ignore=[]):
        """Return a list of symbols of unique elements.

        Parameters:

        ignore: list
            list of symbols to ignore in finding unique elements.
        """
        all_unique = set([a.symbol for a in self.atoms])
        return list(all_unique - set(ignore))

    def single_element_sites(self, allowed_elements):
        """
        Return a list of sites that can only be occupied by a single element
        according to allowed_elements.

        Parameters:

        allowed_elements: list
            List of allowed elements on each site. `allowed_elements` takes the
            same format/style as `basis_elements` in setting (i.e., a nested
            list with each sublist containing a list of elements in a basis).
            It is assumed that all of the first elements in each group is
            present in self.atoms. For example, if `allowed_elements` is
            [['Au', 'Ag' 'X], ['Cu', 'X']] it means that all sites of the
            `self.atoms` must be occupied by either Au or Cu. All of the
            original sites occupied by Au can be occupied by Ag or X in cluster
            expansion, and all of the original sites occupied by Cu can be
            occupied by Cu or X in the cluster expansion.
        """
        single_site_symb = [x[0] for x in allowed_elements if len(x) == 1]
        single_sites = []
        for atom in self.atoms:
            if atom.symbol in single_site_symb:
                single_sites.append(atom.index)
        return single_sites

    def tag_indices_of_corresponding_atom(self, ref_atoms):
        """
        Tag `self.atoms` with the indices of its corresponding atom (equivalent
        position) in `ref_atoms` when the positions are wrapped.


        Parameters:

        ref_atoms: Atoms object
        """
        pos = self.atoms.get_positions()
        wrapped_pos = wrap_positions(pos, ref_atoms.get_cell())
        tree = KDTree(ref_atoms.get_positions())

        dists, indices = tree.query(wrapped_pos)

        if not np.allclose(dists, 0.0):
            raise ValueError("Not all sites has a corresponding atom in the "
                             "passed Atoms object")

        for atom, tag in zip(self.atoms, indices.tolist()):
            atom.tag = tag

    def group_indices_by_trans_symmetry(self, prim_cell):
        """Group indices by translational symmetry.


        Parameters:

        prim_cell: Atoms object
            ASE Atoms object that corresponds to the primitive cell of
            self.atoms
        """
        indices = [a.index for a in prim_cell]
        ref_indx = indices[0]
        # Group all the indices together if its atomic number and position
        # sequences are same
        indx_by_equiv = []
        shifted = wrap_and_sort_by_position(prim_cell.copy())
        an = shifted.get_atomic_numbers()   # atomic number
        pos = shifted.get_positions()       # positions

        temp = [[indices[0]]]
        equiv_group_an = [an]
        equiv_group_pos = [pos]
        for indx in indices[1:]:
            vec = prim_cell.get_distance(indx, ref_indx, vector=True)
            shifted = prim_cell.copy()
            shifted.translate(vec)
            shifted = wrap_and_sort_by_position(shifted)
            an = shifted.get_atomic_numbers()
            pos = shifted.get_positions()

            for equiv_group in range(len(temp)):
                if (an == equiv_group_an[equiv_group]).all() and\
                        np.allclose(pos, equiv_group_pos[equiv_group]):
                    temp[equiv_group].append(indx)
                    break
                else:
                    if equiv_group == len(temp) - 1:
                        temp.append([indx])
                        equiv_group_an.append(an)
                        equiv_group_pos.append(pos)

        for equiv_group in temp:
            indx_by_equiv.append(equiv_group)

        # Now we have found the translational symmetry group of all the atoms
        # in the unit cell, now put all the indices of self.atoms into the
        # matrix based on the tag
        indx_by_equiv_all_atoms = [[] for _ in range(len(indx_by_equiv))]
        symm_group_of_tag = [-1 for _ in range(len(prim_cell))]
        for gr_id, group in enumerate(indx_by_equiv):
            for item in group:
                symm_group_of_tag[item] = gr_id

        for atom in self.atoms:
            symm_gr = symm_group_of_tag[atom.tag]
            indx_by_equiv_all_atoms[symm_gr].append(atom.index)
        return indx_by_equiv_all_atoms

    def corresponding_indices(self, indices, supercell):
        """
        Find the indices in supercell that correspond to the ones in
        self.atoms

        Parameters:

        indices: list of int
            Indices in self.atoms

        supercell: Atoms object
            Supercell object where we want to find the indices
            corresponding to the position in self.atoms
        """
        supercell_indices = []
        sc_pos = supercell.get_positions()
        wrapped_sc_pos = wrap_positions(sc_pos, self.atoms.get_cell())

        dist_to_origin = np.sum(sc_pos**2, axis=1)
        for indx in indices:
            pos = self.atoms[indx].position
            dist = wrapped_sc_pos - pos
            lengths_sq = np.sum(dist**2, axis=1)
            candidates = np.nonzero(lengths_sq < 1E-6)[0].tolist()

            # Pick reference index that is closest the origin of the
            # supercell
            temp_indx = np.argmin(dist_to_origin[candidates])
            supercell_indices.append(candidates[temp_indx])
        return supercell_indices

    def _singlar_cell(self, integer_matrix):
        new_cell = integer_matrix @ self.atoms.cell
        cond = np.linalg.cond(new_cell)
        return cond > 1E10

    def close_to_cubic_supercell(self, zero_cutoff=0.1):
        """
        Create a close to cubic supercell.

        Parameters:

        zero_cutoff: float
            Value below this value will be considered as zero when the
            scaling factor is computed
        """
        cell = self.atoms.get_cell()
        a = np.linalg.det(cell)**(1.0/3.0)
        inv_cell = np.linalg.inv(cell)
        scale = 1.0/inv_cell[np.abs(inv_cell)*a > zero_cutoff]
        scale = np.round(scale).astype(np.int32)
        min_gcd = min([gcd(scale[0], scale[i]) for i in range(len(scale))])

        if min_gcd == 0:
            min_gcd = 1
        scale = np.true_divide(scale, min_gcd)
        scale = min_gcd*np.max(scale)
        integer_matrix = np.round(inv_cell*scale).astype(np.int32)

        counter = 0
        while self._singlar_cell(integer_matrix) and counter < 5:
            row = np.random.randint(low=0, high=3)
            col = np.random.randint(low=0, high=3)
            integer_matrix[row, col] += 1
            counter += 1

        if counter >= 5:
            fname = 'failed_active_template.xyz'
            write(fname, self.atoms)
            raise ValueError("Could not generate a cubic template!"
                             "The activate template is stored in {}"
                             "".format(fname))

        if np.linalg.det(integer_matrix) < 0:
            integer_matrix *= -1

        sc = make_supercell(self.atoms, integer_matrix)
        sc = wrap_and_sort_by_position(sc)

        # We need to tag the atoms
        sc_pos = sc.get_positions()
        sc_pos = wrap_positions(sc_pos, self.atoms.get_cell())

        tree = KDTree(self.atoms.get_positions())
        dists, tags = tree.query(sc_pos)
        assert np.allclose(dists, 0.0)
        for i, tag in enumerate(tags):
            sc[i].tag = tag
        return sc
